@@ -1,7 +1,9 @@
 use desert::{ToBytes,FromBytes,CountBytes};
 use crate::{Scalar,Point,Value,Coord,Error,EyrosErrorKind,Overlap,RA,Root,
   query::QStream, tree_file::TreeFile, SetupFields};
-use async_std::{sync::{Arc,Mutex}};
+use async_std::{sync::{Arc,Mutex},channel};
+#[cfg(not(feature="wasm"))] use async_std::task::spawn;
+#[cfg(feature="wasm")] use async_std::task::{spawn_local as spawn};
 use crate::unfold::unfold;
 use std::collections::{HashMap,HashSet,VecDeque};
 use futures::future::join_all;
@@ -504,108 +506,98 @@ macro_rules! impl_tree {
         }
         (rows,refs)
       }
-      fn query<S>(&mut self, trees: Arc<Mutex<TreeFile<S,Self,($(Coord<$T>),+),V>>>,
+      fn query<S>(&mut self, trees: Arc<TreeFile<S,Self,($(Coord<$T>),+),V>>,
         bbox: &(($($T),+),($($T),+)), fields: Arc<SetupFields>,
         root_index: usize, root_id: TreeId,
       ) -> QStream<($(Coord<$T>),+),V> where S: RA {
-        let istate = (
-          bbox.clone(),
-          vec![], // queue
-          vec![(0usize,Arc::clone(&self.root))], // cursors
-          vec![], // refs
-          Arc::clone(&trees),
-          Arc::clone(&fields)
-        );
-        Box::new(unfold(istate, async move |mut state| {
-          let bbox = &state.0;
-          let queue = &mut state.1;
-          let cursors = &mut state.2;
-          let refs = &mut state.3;
-          let trees = &mut state.4;
-          let fields = &mut state.5;
-          loop {
-            if let Some(q) = queue.pop() {
-              return Some((Ok(q),state));
-            }
-            if cursors.is_empty() && !refs.is_empty() {
-              match Arc::clone(trees).lock().await.get(&refs.pop().unwrap()).await {
-                Err(e) => return Some((Err(e.into()),state)),
-                Ok(t) => cursors.push((0usize,Arc::clone(&t.lock().await.root))),
-              };
-              continue;
-            } else if cursors.is_empty() {
-              if let Err(e) = fields.log(&format!["query end root_index={} root_id={}",
-              root_index, root_id]).await {
-                return Some((Err(e.into()),state));
-              }
-              return None;
-            }
-            let (level,c) = cursors.pop().unwrap();
-            match c.as_ref() {
-              $Node::Branch(branch) => {
-                match level % $dim {
-                  $($i => {
-                    let pivots = branch.pivots.$i.as_ref().unwrap();
+        let nproc = std::thread::available_concurrency().map(|n| n.get()).unwrap_or(1);
+        let (refs_sender,refs_receiver) = channel::unbounded();
+        let (queue_sender,queue_receiver) = channel::bounded::<
+          Result<(Vec<(($(Coord<$T>),+),V)>,Vec<TreeId>),Error>
+        >(nproc);
 
-                    {
-                      let mut matching: u32 = 0;
-                      if &(bbox.0).$i <= pivots.first().unwrap() {
-                        matching |= (1<<0);
-                      }
-                      let ranges = pivots.iter().zip(pivots.iter().skip(1));
-                      for (i,(start,end)) in ranges.enumerate() {
-                        if intersect_iv(start, end, &(bbox.0).$i, &(bbox.1).$i) {
-                          matching |= (1<<i);
-                          matching |= (1<<(i+1));
-                        }
-                      }
-                      if &(bbox.1).$i >= pivots.last().unwrap() {
-                        matching |= (1<<(pivots.len()-1));
-                      }
-                      for (bitfield,b) in branch.intersections.iter() {
-                        if (matching & bitfield) > 0 {
-                          cursors.push((level+1,Arc::clone(b)));
-                        }
-                      }
-                    }
-
-                    {
-                      let xs = &branch.nodes;
-                      let ranges = pivots.iter().zip(pivots.iter().skip(1));
-                      if &(bbox.0).$i <= pivots.first().unwrap() {
-                        cursors.push((level+1,Arc::clone(xs.first().unwrap())));
-                      }
-                      for ((start,end),b) in ranges.zip(xs.iter().skip(1)) {
-                        if intersect_iv(start, end, &(bbox.0).$i, &(bbox.1).$i) {
-                          cursors.push((level+1,Arc::clone(b)));
-                        }
-                      }
-                      if &(bbox.1).$i >= pivots.last().unwrap() {
-                        cursors.push((level+1,Arc::clone(xs.last().unwrap())));
-                      }
-                    }
-                  }),+
-                  _ => panic!["unexpected level modulo dimension"]
+        // todo: factor out locking
+        for _ in 0..nproc {
+          let refs_r = refs_receiver.clone();
+          let queue_s = queue_sender.clone();
+          let bbox_c = bbox.clone();
+          let trees_c = trees.clone();
+          spawn(async move {
+            while let Ok(r) = refs_r.recv().await {
+              match trees_c.get(&r).await {
+                Err(e) => queue_s.send(Err(e.into())).await.unwrap(),
+                Ok(t) => {
+                  let (rows,tr_refs) = t.lock().await.query_local(&bbox_c);
+                  let refs = tr_refs.iter().map(|r| r.id).collect::<Vec<TreeId>>();
+                  queue_s.send(Ok((rows,refs))).await.unwrap();
                 }
-              },
-              $Node::Data(data,rs) => {
-                queue.extend(data.iter()
-                  .filter(|pv| {
-                    true $(&& intersect_coord(&(pv.0).$i, &(bbox.0).$i, &(bbox.1).$i))+
-                  })
-                  .map(|pv| { pv.clone() })
-                  .collect::<Vec<_>>()
-                );
-                refs.extend(rs.iter()
-                  .filter(|r| {
-                    true $(&& intersect_coord(&r.bounds.$i, &(bbox.0).$i, &(bbox.1).$i))+
-                  })
-                  .map(|r| { r.id })
-                  .collect::<Vec<TreeId>>()
-                );
-              },
+              };
+            }
+          });
+        }
+        struct QState<$($T: Scalar),+, V: Value> {
+          refs: VecDeque<TreeId>,
+          results: VecDeque<(($(Coord<$T>),+),V)>,
+          active: usize,
+          queue_r: channel::Receiver<Result<(Vec<(($(Coord<$T>),+),V)>,Vec<TreeId>),Error>>,
+          queue_s: channel::Sender<Result<(Vec<(($(Coord<$T>),+),V)>,Vec<TreeId>),Error>>,
+          refs_s: channel::Sender<TreeId>,
+          fields: Arc<SetupFields>,
+        }
+        let istate = {
+          let (v_results,v_refs) = self.query_local(bbox);
+          let mut refs = VecDeque::with_capacity(v_refs.len());
+          let mut results = VecDeque::with_capacity(v_results.len());
+          for r in v_results { results.push_back(r); }
+          for r in v_refs { refs.push_back(r.id); }
+          QState {
+            refs,
+            results,
+            active: 0,
+            queue_r: queue_receiver.clone(),
+            queue_s: queue_sender.clone(),
+            refs_s: refs_sender.clone(),
+            fields: fields.clone(),
+          }
+        };
+        Box::new(unfold(istate, async move |mut state| {
+          loop {
+            if let Some(res) = state.results.pop_front() {
+              return Some((Ok(res),state));
+            } else if state.active > 0 {
+              match state.queue_r.recv().await.unwrap() {
+                Ok((v_results,v_refs)) => {
+                  state.active -= 1;
+                  state.results.reserve(v_results.len());
+                  state.refs.reserve(v_refs.len());
+                  for r in v_results { state.results.push_back(r); }
+                  for r in v_refs { state.refs.push_back(r); }
+                },
+                Err(e) => {
+                  state.active -= 1;
+                  return Some((Err(e.into()),state));
+                },
+              }
+            } else if !state.refs.is_empty() {
+              for _ in 0..nproc {
+                if let Some(r) = state.refs.pop_front() {
+                  state.active += 1;
+                  state.refs_s.send(r).await.unwrap();
+                } else {
+                  break;
+                }
+              }
+            } else {
+              break;
             }
           }
+          state.refs_s.close();
+          state.queue_s.close();
+          if let Err(e) = state.fields.log(&format!["query end root_index={} root_id={}",
+          root_index, root_id]).await {
+            return Some((Err(e.into()),state));
+          }
+          None
         }))
       }
       async fn remove<S>(&mut self, xids: Arc<Mutex<HashMap<V::Id,($(Coord<$T>),+)>>>)
@@ -722,7 +714,7 @@ where P: Point, V: Value {
   fn list(&mut self) -> (Vec<(P,V)>,Vec<TreeRef<P>>);
   fn list_refs(&mut self) -> Vec<TreeRef<P>>;
   fn query_local(&mut self, bbox: &P::Bounds) -> (Vec<(P,V)>,Vec<TreeRef<P>>);
-  fn query<S>(&mut self, trees: Arc<Mutex<TreeFile<S,Self,P,V>>>, bbox: &P::Bounds,
+  fn query<S>(&mut self, trees: Arc<TreeFile<S,Self,P,V>>, bbox: &P::Bounds,
     fields: Arc<SetupFields>, root_index: usize, root_id: TreeId) -> QStream<P,V> where S: RA;
   async fn remove<S>(&mut self, ids: Arc<Mutex<HashMap<V::Id,P>>>)
     -> (Option<(Vec<(P,V)>,Vec<TreeRef<P>>)>,Vec<TreeId>) where S: RA;
@@ -735,7 +727,7 @@ where P: Point, V: Value, T: Tree<P,V>, S: RA {
   pub deletes: Arc<Vec<(P,V::Id)>>,
   pub inputs: Arc<Vec<TreeRef<P>>>,
   pub roots: Vec<Root<P>>,
-  pub trees: Arc<Mutex<TreeFile<S,T,P,V>>>,
+  pub trees: Arc<TreeFile<S,T,P,V>>,
   pub next_tree: &'a mut TreeId,
   pub rebuild_depth: usize,
   pub error_if_missing: bool,
@@ -755,10 +747,9 @@ where P: Point, V: Value, T: Tree<P,V>, S: RA {
     l_refs.extend_from_slice(&self.inputs);
 
     for _ in 0..self.rebuild_depth {
-      let mut trees = self.trees.lock().await;
       let mut n_refs = vec![];
       for r in l_refs.iter() {
-        let (list,xrefs) = trees.get(&r.id).await?.lock().await.list();
+        let (list,xrefs) = self.trees.get(&r.id).await?.lock().await.list();
         lists.push(list);
         n_refs.extend(xrefs);
         rm_trees.push(r.id);
@@ -807,15 +798,15 @@ where P: Point, V: Value, T: Tree<P,V>, S: RA {
       if ro.is_none() { continue }
       let r = ro.as_ref().unwrap();
       // TODO: remove the delete when found
-      let trees = Arc::clone(&self.trees);
-      let xids = Arc::clone(&ids);
+      let trees = self.trees.clone();
+      let xids = ids.clone();
       let id = r.id;
       let xfields = Arc::clone(&fields);
       work.push(async move {
         let mut refs = vec![id];
         while !refs.is_empty() {
           let r = refs.pop().unwrap();
-          let tm = trees.lock().await.get(&r).await?;
+          let tm = trees.get(&r).await?;
           let mut t = tm.lock().await;
           let (built,nrefs) = t.remove::<S>(
             Arc::clone(&xids),
@@ -831,7 +822,7 @@ where P: Point, V: Value, T: Tree<P,V>, S: RA {
             }).collect::<Vec<_>>());
             let mut next_tree = r;
             if rows.is_empty() {
-              trees.lock().await.put(&r, Arc::new(Mutex::new(T::empty()))).await?;
+              trees.put(&r, Arc::new(Mutex::new(T::empty()))).await?;
             } else {
               let (tr, create_trees) = T::build(
                 Arc::clone(&xfields),
@@ -846,7 +837,7 @@ where P: Point, V: Value, T: Tree<P,V>, S: RA {
               ];
               assert![create_trees.len() == 1, "unexpected external sub-trees during remove()"];
               for (r,t) in create_trees {
-                trees.lock().await.put(&r, t).await?;
+                trees.put(&r, t).await?;
               }
             }
           }
